/* 
 * TiVo's "fun" function counter trap madness.  We insert traps
 * into the function prologue, the trap opcode is encoded with
 * a function number.  We use that number as an offset into
 * an area of memory where we track the number of times that
 * function was called.
 *
 * Copyright (C) 2004 TiVo Inc. All Rights Reserved.
 * 
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernprof.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/proc_fs.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/smp.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/module.h>
#include <linux/compiler.h>

#include <asm/uaccess.h>

#define FUNCOUNT_NUMFUNC (100000)
#define FUNCOUNT_BYTES_PER_FUNC (sizeof(unsigned int))
#define FUNCOUNT_SIZE (FUNCOUNT_NUMFUNC * FUNCOUNT_BYTES_PER_FUNC)

char* funcount_buffer = NULL; /* Holds pointer to funcount buffer in the
			  	case of funcount being enabled, otherwise
			  	holds NULL */
unsigned int funorder_count = 0; /* Holds value of current function order value
			   if we're in funorder mode instead of
			   funcount mode */

static char* funcount_buffer_ptr; /* Holds pointer to funcount buffer */

static ssize_t funcount_enable_read(struct file *filp, char *user_buf,
				   size_t count, loff_t *ppos) {
	if(*ppos > 0)
		return 0;
	if(count < 2)
		return 0;
	if(funcount_buffer != NULL) {
		if((unsigned long)funcount_buffer & 1) /* funorder mode */
			copy_to_user(user_buf, "2\n", 2);
		else /* funcount mode */
			copy_to_user(user_buf, "1\n", 2);
	} else {
		copy_to_user(user_buf, "0\n", 2);
	}
	*ppos += 2;
	return 2;
}

static ssize_t funcount_enable_write(struct file *filp, const char *user_buf,
			      size_t count, loff_t *data) {
	char val[2];
	if(count < 2)
		return -EINVAL;
	copy_from_user(&val, user_buf, 2);
	if(val[0] != '0' && val[0] != 0) { /* start case */
		if(funcount_buffer != NULL) {
			funcount_buffer=NULL;
		}
		if(funcount_buffer_ptr != NULL) {
			memset(funcount_buffer_ptr, 0, FUNCOUNT_SIZE);	
			funcount_buffer=funcount_buffer_ptr;
			if(val[0] == '2') {
				funcount_buffer = (void*)((unsigned int)funcount_buffer | 1); 
				funorder_count=0; /* ASM increments before store */
			}
		}
	} else { /* stop case */
		funcount_buffer=NULL;
	}
	return 2;
}

static ssize_t funcount_data_read(struct file *filp, char* user_buf,
				   size_t count, loff_t *ppos) {
	if(*ppos > FUNCOUNT_SIZE)
		return 0;
	if(count > FUNCOUNT_SIZE - *ppos)
		count = FUNCOUNT_SIZE - *ppos;
	if(funcount_buffer != NULL)
		funcount_buffer=NULL;
	if(copy_to_user(user_buf, funcount_buffer_ptr + *ppos, count))
		return -EFAULT;
	*ppos += count;
	return count;
}

static struct file_operations proc_funcount_enable_fops = {
	read: funcount_enable_read,
	write:funcount_enable_write,
};

static struct file_operations proc_funcount_data_fops = {
	read: funcount_data_read,
};

static int create_proc_entries(void) {
	struct proc_dir_entry *entena, *entdata, *entdir;
	if((entdir = proc_mkdir("funcount", NULL)) == NULL) {
		printk(KERN_ERR "funcount: unable to create /proc entry!\n");
		return -ENODEV;
	}
	if((entena = create_proc_entry("enabled", 0, entdir)) != NULL) {
		entena->size = 2;
		entena->proc_fops = &proc_funcount_enable_fops;
	} else {
		remove_proc_entry("funcount", NULL);
		printk(KERN_ERR "funcount: unable to create /proc entry!\n");
		return -ENODEV;
	}
	if((entdata = create_proc_entry("data", 0, entdir)) != NULL) {
		entdata->size = FUNCOUNT_SIZE;
		entdata->proc_fops = &proc_funcount_data_fops;
	} else {
		remove_proc_entry("enabled", entdir);
		remove_proc_entry("funcount", NULL);
		printk(KERN_ERR "funcount: unable to create /proc entry!\n");
		return -ENODEV;
	}
	printk("funcount enabled, see /proc/funcount/ to start/stop/capture function counting\n");
	return 0;
}

static int startup_funcount=0;

enum {
	FUNCOUNT_STMODE_INITONLY = 1,
	FUNCOUNT_STMODE_COUNTS,
	FUNCOUNT_STMODE_ORDER,
};

static int __init funcount_init(void) {
	if(startup_funcount) {
		funcount_buffer_ptr = vmalloc(FUNCOUNT_SIZE);
		if(funcount_buffer_ptr == NULL) {
			printk("funcount disabled, unable to allocate space for count location\n");
			return 0;
		}

		memset(funcount_buffer_ptr, 0, FUNCOUNT_SIZE);
		create_proc_entries();
		switch(startup_funcount) {
		case FUNCOUNT_STMODE_COUNTS:
			funcount_buffer = funcount_buffer_ptr;
			printk("funcount startup in count mode\n");
			break;
		case FUNCOUNT_STMODE_ORDER:
			funcount_buffer = (void*)((unsigned int)funcount_buffer_ptr | 1);
			printk("funcount startup in order mode\n");
			break;
		}
		return 0;
	}
	printk("funcount disabled\n");
	return 0;
}

static int __init funcount_setup_init(char *str) {
	if(strncmp(str, "true", 4) == 0) {
		startup_funcount=FUNCOUNT_STMODE_INITONLY;
		return 0;
	}
	if(strncmp(str, "st", 2) == 0) {
		if(strncmp(str, "stcount", 7) == 0) {
			startup_funcount=FUNCOUNT_STMODE_COUNTS;
		} else {
			startup_funcount=FUNCOUNT_STMODE_ORDER;
		}
		return 0;
	}
	printk("funcount disabled, specifiy funcount=true to enable\n");
	return 0;
}
__setup("funcount=", funcount_setup_init);
module_init(funcount_init);
