/*
 * Common code for monotonic time on MIPS-based TiVo systems
 *
 * Copyright (C) 2002-2005 TiVo Inc. All Rights Reserved.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/param.h>
#include <linux/proc_fs.h>
#include <linux/interrupt.h>
#include <linux/tivo-monotime.h>
#ifdef CONFIG_SW_WATCHDOG
#include <linux/sw_watchdog.h>
#endif

#include <asm/mipsregs.h>
#include <asm/system.h>

extern int get_counter_irq(void);

#ifdef CONFIG_SW_WATCHDOG
unsigned long _gbl_sti_cli_flags = 0;
EXPORT_SYMBOL(_gbl_sti_cli_flags);
unsigned int sw_watchdog_state = SW_WATCHDOG_DISABLED;
int sw_watchdog_fl_irq=0; /* Current first level IRQ */
#endif

volatile unsigned int tivo_time_hi;

/*
 * tivo_time_getcount implements a guaranteed monotonic time counter,
 * using the MIPS cycle counter and an interrupt-driven wrap counter
 *
 * Note that this operation is now fast-pathed in the trap exception
 * handler. This code will only be called if a trap instruction is
 * placed inside a branch delay slot, which the fast-path will not
 * detect.
 */
void tivo_time_getcount(struct pt_regs *regs)
{
	unsigned int hi1, hi2, lo;

	do {
		hi1 = tivo_time_hi;
		lo = read_32bit_cp0_register(CP0_COUNT);
		hi2 = tivo_time_hi;
	} while (hi1 != hi2);
	regs->regs[2] = hi1;		// high order bits -> v0
	regs->regs[3] = lo;		// low order bits  -> v1
}

/*
 * same monotonic count, for kernel functions
 */
void tivo_monotonic_get_current(monotonic_t *t)
{
	u32 hi1, hi2, lo;

	do {
		hi1 = tivo_time_hi;
		lo = read_32bit_cp0_register(CP0_COUNT);
		hi2 = tivo_time_hi;
	} while (hi1 != hi2);
	t->count = ((u64)hi1 << 32) | lo;
}
EXPORT_SYMBOL(tivo_monotonic_get_current);

static unsigned int prof_interval;
static unsigned int prof_compare;

static void tivo_time_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
#ifdef CONFIG_SW_WATCHDOG
	if(local_irq_count(smp_processor_id()) > 1) {
		printk(KERN_DEBUG "Watchdog re-entry %d\n", sw_watchdog_fl_irq);
	}
	if(sw_watchdog_fire()) {
		printk("Watchdog Fired!\n");
		*(unsigned int*)0=0;
	}
	sw_watchdog_reset();
#endif
	
	if (prof_interval == 0) {
		tivo_time_hi++;
		write_32bit_cp0_register(CP0_COMPARE, 0);
	} else {
		unsigned int next_compare;

		next_compare = read_32bit_cp0_register(CP0_COMPARE);

		do {
			if (next_compare == 0) {
				tivo_time_hi++;
			}

			if (prof_compare == next_compare) {
				prof_compare += prof_interval;
				if (prof_compare < prof_interval) {
					// Addition wrapped, set the next
					// interrupt to catch counter wrap
					next_compare = 0;
				} else {
					// Addition did not wrap, set the next
					// interrupt to our next profiling tick
					next_compare = prof_compare;
				}
			} else {
				next_compare = prof_compare;
			}

			// Writing the CP0_COMPARE register will acknowledege
			// the interrupt. We need to account for the case that
			// the new compare value we are writing has just
			// passed, and treat that as though the interrupt
			// triggered.
			write_32bit_cp0_register(CP0_COMPARE, next_compare);
			barrier();
			// Logically, reading the count after setting the
			// compare should make a simple compare < count be
			// sufficient check for race. In practice, however,
			// the new compare value does not take effect until
			// several ticks later, so we allow for that.
		} while ((signed)(next_compare - read_32bit_cp0_register(CP0_COUNT) - 6) < 0);
	}
}

unsigned long tivo_counts_per_256usec;
EXPORT_SYMBOL(tivo_counts_per_256usec);

static int counter_irq;

static int prof_interval_write_proc(struct file * pFile, const char * pBuffer,
                                    unsigned long input_length, void *data )
{
	int interval;

	interval = (int)simple_strtol(pBuffer, NULL, 0);

	// Negative value not supported for now, may be used for something
	// in the future.
	//
	if (interval < 0) {
		return -EINVAL;
	}

	// Don't allow too small an interval, to avoid getting stuck in a
	// loop of just servicing the interrupt over and over again.
	//
	if (interval > 0 && interval < 2500) {
		interval = 2500;
	}

	// Uhg. Can't just disable interrupts around this while thing,
	// because if we did, the write to CP0_COMPARE could cause us to
	// miss a wrap interrupt, so we handle the potential race conditions
	// as best as we can.
	//
	// First set the interval to 0, in a way that should not miss
	// any wrap interrupts.
	//
	write_32bit_cp0_register(CP0_COMPARE, 0);
	barrier();
	prof_interval = 0;
	barrier();
	write_32bit_cp0_register(CP0_COMPARE, 0);

	// Then if we're setting the interval to a non-0 value, only
	// write CP0_COMPARE if we're not going to get the wrap interrupt
	// before the next profile tick. Besides being what the interrupt
	// handler expects, that should be sufficient condition to avoid
	// acknowledging an interrupt that we didn't mean to acknowledge.
	//
	if (interval != 0) {
		unsigned int next_compare;
		unsigned int uinterval = (unsigned int)interval;

		cli();

		next_compare = read_32bit_cp0_register(CP0_COUNT) + uinterval;

		if (next_compare > uinterval) {
			write_32bit_cp0_register(CP0_COMPARE, next_compare);
		} else {
			// Leave CP0_COMPARE at 0.
		}

		prof_interval = uinterval;
		prof_compare = next_compare;

		sti();
	}

	return input_length;
}

static int prof_interval_read_proc(char * pBuffer, char **start, off_t offset,
                                   int length, int *eof, void *data)
{
	int len = 0;

	len = sprintf(pBuffer, "IRQ: %d\nInterval: %d\n",
	              counter_irq, prof_interval) - offset;
	if (len < length) {
		*eof = 1;
		if (len <= 0) return 0;
	} else {
		len = length;
	}

	*start = pBuffer + offset;
	return len;
}
int tivo_monotonic_initialized = 0;
static int __init tivo_monotonic_init(void)
{
	int rv;
	struct proc_dir_entry *entry;

	counter_irq = get_counter_irq();

	write_32bit_cp0_register(CP0_COMPARE, 0);
	rv = request_irq(counter_irq, tivo_time_interrupt, SA_INTERRUPT,
	                 "counter", NULL);
	if (rv) {
		return rv;
	}

	entry = create_proc_entry("prof_interval", 0, &proc_root);
	if (entry) {
		entry->write_proc = prof_interval_write_proc;
		entry->read_proc = prof_interval_read_proc;
	} else {
		return -ENOMEM;
	}
    tivo_monotonic_initialized = 1; 
	return 0;
}

module_init(tivo_monotonic_init);

