///////////////////////////////////////////////////////////////////////////
//
// File: bcmenet.c
//
// Copyright (c) 2005, TiVo Inc.
//
///////////////////////////////////////////////////////////////////////////

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <asm/io.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/tqueue.h>

#include <linux/mii.h>
#include <linux/crc32.h>

#include <asm/bcm7038.h>

#include <net/realign.h>

#include "bcmenet.h"


MODULE_DESCRIPTION( "TiVo BCM ENET Ethernet driver" );
MODULE_LICENSE( "GPL" );

static int debug = 0;
MODULE_PARM( debug, "i");
MODULE_PARM_DESC( debug, "flag to control the debug printk level");

#ifdef MODULE
static unsigned char *mac= NULL;
MODULE_PARM(mac, "s");
MODULE_PARM_DESC( mac,"mac address to be configured on the port");
#else
static unsigned char mac[256];
#endif /* MODULE */

static int gMaxRxDescs = MAX_RX_DESCS;
MODULE_PARM(gMaxRxDescs, "i");
MODULE_PARM_DESC( gMaxRxDescs,"maximum number of receive dma descriptors [2-64]");


static int gNumberOfOverflows = 0;


#define PRINTK( x... )                                              \
    do {                                                            \
        if (debug != 0) {                                           \
            printk(x);                                              \
        }                                                           \
    }                                                               \
    while (0)   


#define register_write(reg, value)                                  \
    do {                                                            \
        *__longreg(reg) = (value);                                  \
        barrier();                                                  \
    }while(0)

#define register_read(var, reg)                                     \
    do {                                                            \
        (var) = 0;                                                  \
        barrier();                                                  \
        (var) = *(__longreg(reg));                                  \
        barrier();                                                  \
    }while(0)

#define register_set_bits(reg,  value)                              \
    do {                                                            \
        *__longreg(reg) |= (value);                                 \
        barrier();                                                  \
    }while(0)

#define register_set(reg, mask, shift, value)                       \
    do {                                                            \
        *__longreg(reg) &=  ~(mask);                                \
        *__longreg(reg) |= (((value) << (shift)));                  \
        barrier();                                                  \
    }while(0)

#define register_clear(reg, mask)                                   \
    do {                                                            \
        *__longreg(reg) &= ~(mask);                                 \
        barrier();                                                  \
    }while(0)


#define DMA_DESC_FILL_BUFFER_PARAMS(desc, buf, len)                 \
    do {                                                            \
        register_write(&((desc)->length_and_status),                \
                     (((len) & 0xFFFUL) << DMA_DESC_BLENGTH_SHIFT));\
        barrier();                                                  \
        if (((desc) == &ENET_BRIDGE_TX_DESC_TABLE[MAX_TX_DESCS-1])||\
            ((desc) == &ENET_BRIDGE_RX_DESC_TABLE[gMaxRxDescs-1])){ \
            register_set_bits(&((desc)->length_and_status),         \
                            DMA_DESC_WRAP);                         \
            barrier();                                              \
        }                                                           \
        register_write(&((desc)->buffer),                           \
                       ((unsigned long)(buf) & 0xffffffffUL));      \
        barrier();                                                  \
    }while (0)

#define is_desc_free(desc) (((desc)->length_and_status & DMA_DESC_OWN) ? 0:1)

enum enet_worker_state  {
    ENET_WORKER_START_RX_DMA  = 0  ,
    ENET_WORKER_FINISH_RX_DMA      ,

    /* the end */
    ENET_WORKER_STATE_END
};

#ifndef CONFIG_TIVO_REALIGN_ETHERNET
#define COPY_PACKET
#endif

typedef struct {
    struct list_head        link;
    volatile dma_desc_t *   descriptor;
    struct sk_buff *        skb;
#ifdef COPY_PACKET
    char *                  rxBufferHead;
    char *                  rxBuffer;
#endif /* COPY_PACKET */
}enet_desc_elem_t;


typedef struct {
    struct list_head        link;
    enum enet_worker_state  state;
    enet_desc_elem_t *      elem;
    unsigned int            sn;
    int                     go;
    unsigned long           priv;
}enet_worker_t;

typedef struct {
    struct net_device*              net;
    struct net_device_stats         stats;

    spinlock_t                      lock;

    struct tq_struct                enet_worker_task;
    struct tq_struct                enet_receiver_task;
    struct timer_list               enet_timer;
    struct list_head                workers;
    unsigned long                   workerSN;

    enet_worker_t *                 receiver;

    enet_desc_elem_t *              rx_descs[MAX_RX_DESCS];
    
    struct sk_buff *                tx_descs[MAX_TX_DESCS];
    int                             curTxIdx;
    atomic_t                        reclaimTxDescs;

    unsigned char                   mac_address[6];
    unsigned long                   pm_reg_entry[MAX_PM_REGS];
    int                             linkState;

}enet_device_t;

enum mii_config {
    MII_100MBIT     = 0x0001,
    MII_FULLDUPLEX  = 0x0002,
    MII_AUTONEG     = 0x0004,

    /* end */
    MII_END
};

static enet_device_t gContext;

static void enet_task (void *data);
static void enet_receiver_task_func (void *data);
static void enet_timer_func(unsigned long data);
static int  enet_run_workers (void);
static void enet_worker_start_rx_dma(enet_worker_t *worker);
static void enet_worker_finish_rx_dma(enet_worker_t *worker);
static void enet_worker_next_state(enet_worker_t *worker, 
                                   enum enet_worker_state nextState);

static int      enet_init(void);
static int      enet_create_and_assign_rx_skbs(void);
static int      enet_udma_init(enet_device_t *context);
static int      enet_emac_init(enet_device_t *context);
static int      parse_mac_address_parameter(enet_device_t *context);
static int      pm_reg_write(int index, unsigned char * addr, int valid);
static int      write_mac_address(struct net_device* net);
static int      disable_emac(void);
    
static int      enet_set_mac_address (struct net_device *net, void *p);
static int      enet_open (struct net_device *net);
static int      enet_stop (struct net_device *net);
static void     enet_net_timeout (struct net_device *net);
static void     enet_set_mcast_list (struct net_device *net);
static struct   net_device_stats * enet_get_stats(struct net_device *net);
static int      enet_hard_start_xmit(struct sk_buff *skb, 
                                     struct net_device *net);
static void     enet_isr(int irq, void *dev_id, struct pt_regs *regs);


static enum mii_config mii_get_config(void);
static unsigned long mii_read(unsigned long uPhyAddr, unsigned long uRegAddr); 
static enum mii_config mii_auto_configure(void);
static void mii_write(unsigned long phyAddr, unsigned long regAddr, 
                      unsigned long data);
static void wait_for_mii_interrupt(void);
int is_enet_up(unsigned long uPhyAddr);

static void zero_descriptor_table(volatile dma_desc_t *table, size_t len);
static void zero_enet_bridge(void);
static unsigned long enet_do_packet_receive(unsigned long curIdx);

#ifndef MODULE
static int __init enet_debug_setup(char *str)
{
    get_option(&str, &debug);
    return 1;
}

static int __init enet_mac_addr_setup (char *str)
{
    if (str != NULL) {
        strncpy(mac, str, 256);
    }
    return 1;
}

__setup("dbgenet=", enet_debug_setup);
__setup("mac=", enet_mac_addr_setup);

#endif /* ! MODULE */


#define enet_make_worker_queue_runnable()                               \
    do{                                                                 \
        queue_task(&gContext.enet_worker_task, &tq_immediate);          \
        mark_bh(IMMEDIATE_BH);                                          \
    }while (0)

#define enet_rearm_timer_task()                                         \
    do {                                                                \
        gContext.enet_timer.expires = jiffies + (HZ);                   \
        add_timer(&gContext.enet_timer);                                \
    }while(0)


#define ENET_MII_PHYADDR         1
/* MII Control */
#define EPHY_LOOPBACK                   0x4000
#define EPHY_AUTONEGOTIATE              0x1000
#define EPHY_100MBIT                    0x2000
#define EPHY_FULLDUPLEX                 0x0100
#define EPHY_RESTART_AUTONEGOTIATION    0x0200
/* MII STATUS */
#define EPHY_AUTONEGOTIATION_COMPLETE   0x0020
/* MII Auxiliary Mode 2 */
#define EPHY_ACT_LINK_LED               0x0004


/* 
 * mii_get_config: Return the current MII configuration
 */
static enum mii_config mii_get_config()
{
    unsigned long  val;
    enum mii_config config = 0;

    /* Read the Link Partner Ability */
    val = mii_read(ENET_MII_PHYADDR, 0x05) & 0x0000FFFF;
    if (val & 0x0100) {           /* 100 MB Full-Duplex */
        config = (MII_100MBIT | MII_FULLDUPLEX);
    } else if (val & 0x0080) {    /* 100 MB Half-Duplex */
        config = MII_100MBIT;
    } else if (val & 0x0040) {    /* 10 MB Full-Duplex */
        config = MII_FULLDUPLEX;
    } 

    return config;
}


/*
 * mii_AutoConfigure: Auto-Configure this MII interface
 */
static enum mii_config mii_auto_configure()
{
    int i;
    unsigned long  val;
    enum mii_config config;

    /* enable and restart autonegotiation */
    val = mii_read(ENET_MII_PHYADDR, 0) & 0x0000FFFF;
    val |= (EPHY_RESTART_AUTONEGOTIATION | EPHY_AUTONEGOTIATE);
    mii_write(ENET_MII_PHYADDR, 0x00, val);

    /* wait for it to finish */
    for (i = 0; i < 1000; i++) {
        mdelay(1);
        val = mii_read(ENET_MII_PHYADDR, 0x01);
        if (val & EPHY_AUTONEGOTIATION_COMPLETE) {
            break;
        }
    }

    config = mii_get_config();
    if (val & EPHY_AUTONEGOTIATION_COMPLETE) {
        config |= MII_AUTONEG;
    } 
    mii_write(ENET_MII_PHYADDR, 0x1A, 0x0F00);

    return config;
}

/*
 * mii_write: Write a value to the MII
 */
static void wait_for_mii_interrupt()
{                                                                   
  volatile unsigned long val;                               
  int count = 10000;
  do {                                                            
        register_read(val, ENET_BRIDGE_INT_STATUS);               
        udelay(10);                                               
  }while (!(val & EMAC_STATUS_MII) && --count > 0 );                                 
}


            
static void mii_write(unsigned long phyAddr, unsigned long regAddr, unsigned long data)
{
    unsigned long d;
    d = 0x50020000 | (phyAddr << 23) | (regAddr << 18) | data;
    register_write(__longreg(&(ENET_BRIDGE->mii_data)),d);
    mdelay(1);
    wait_for_mii_interrupt();
    register_set_bits(ENET_BRIDGE_INT_STATUS,EMAC_STATUS_MII);
}

/*
 * mii_read: Read a value from the MII
 */
static unsigned long mii_read(unsigned long uPhyAddr, unsigned long uRegAddr) 
{
    volatile unsigned long val;

    val = 0x60020000 | (uPhyAddr << 23) | (uRegAddr << 18);
    register_write(__longreg(&(ENET_BRIDGE->mii_data)), val);
    mdelay(1);
    wait_for_mii_interrupt();
    register_set_bits(ENET_BRIDGE_INT_STATUS,EMAC_STATUS_MII);
    register_read(val, __longreg(&(ENET_BRIDGE->mii_data)));
    return val;
}

/*
 * soft_reset: Reset the MII
 */
static void soft_reset(unsigned long uPhyAddr) 
{
    unsigned long uData;

    mii_write(uPhyAddr, 0, 0x8000);
    // wait ~10usec
    udelay(10);
    do {
        uData = mii_read(uPhyAddr, 0);
    } while (uData & 0x00008000);

}

/* get ethernet port's status; return nonzero for Link up, 0 for Link down */
int is_enet_up(unsigned long uPhyAddr)
{
    static int linkState = 0;
    static int sem = 0;
    volatile unsigned short value;

    if ( sem != 0 ) {
        return linkState;
    }else {
        sem = 1;
    }

    value = (unsigned short)(mii_read(uPhyAddr,0x1) & 0xffff);
    if (value & 0x20) {
        linkState = 1;
    } else {
        linkState = 0;
    }

    if (linkState != gContext.linkState) {
        gContext.linkState = linkState;
        if (linkState == 0) {
            printk("enet : link down\n");
            netif_carrier_off(gContext.net);
        }else {
            printk("enet : link up\n");
            netif_carrier_on(gContext.net);
        }
    }

    sem = 0;
    return linkState;
}

static void empty_desc_list(void)
{
    int i = 0;
    volatile unsigned long flags;
    spin_lock_irqsave(&gContext.lock, flags);
    for (i = 0 ;  i < MAX_TX_DESCS ; ++i) {
        if (NULL != gContext.tx_descs[i]) {
            dev_kfree_skb_any(gContext.tx_descs[i]);
            gContext.tx_descs[i] = NULL;
        }
    }

    spin_unlock_irqrestore(&gContext.lock, flags);

}


static void enet_cleanup(void)
{
    del_timer(&gContext.enet_timer);
    unregister_netdev(gContext.net);
    kfree(gContext.net);

    free_irq(gContext.net->irq, gContext.net);

    empty_desc_list();

    if (gContext.receiver != NULL) {
        kfree(gContext.receiver);
    }
    gContext.receiver = NULL;

    register_clear(ENET_DMA_CONTROLLER_R5K_IRQ_MSK, 
                   IUDMA_R5K_IRQ_STATUS_IUDMA_MASK);

}

static void zero_descriptor_table(volatile dma_desc_t *table, size_t len)
{
    unsigned int i = 0;
    for (i = 0; i < len ; i++) {
        DMA_DESC_FILL_BUFFER_PARAMS(&(table[i]), 0, 0);
    }
}

static void zero_enet_bridge()
{
    unsigned int i = 0;
    volatile unsigned long *ptr = (volatile unsigned long *)ENET_BRIDGE;
    for (i = 0; i < sizeof(enet_bridge_t); i+=sizeof(unsigned long),++ptr) {
        *ptr = 0;
    }
}

static inline void enet_worker_next_state(enet_worker_t *worker, 
                                          enum enet_worker_state nextState)
{
    worker->state = nextState;
    worker->go    = 1;
}


static int __init enet_init(void)
{
    int ret = 0;
    int i = 0;
    struct net_device *dev = NULL;

#ifdef COPY_PACKET
    PRINTK(KERN_INFO "enet: using COPY_PACKET\n");
#else 
    PRINTK(KERN_INFO "enet: not using COPY_PACKET\n");
#endif    

    zero_enet_bridge();

    PRINTK(KERN_INFO "enet bridge(%d bytes@0x%08X)\n", 
            sizeof(enet_bridge_t), (unsigned int)ENET_BRIDGE);
    PRINTK(KERN_INFO "controller Revision: 0x%04X\n", (unsigned int)*ENET_DMA_CONTROLLER_REVISION);
    /* clear the descriptor tables */
    zero_descriptor_table(ENET_BRIDGE_TX_DESC_TABLE, MAX_TX_DESCS);
    zero_descriptor_table(ENET_BRIDGE_RX_DESC_TABLE, MAX_RX_DESCS);

    /* sw reset the ethernet controller */
    *(SUN_TOP_CONTROL_SW_RESET) |= 0x00000010UL;
    mdelay(200);
    *(SUN_TOP_CONTROL_SW_RESET) &= ~0x00000010UL;

    memset(&gContext, 0, sizeof(enet_device_t));

    /* initialize the udma engine */
    ret = enet_udma_init(&gContext);
    if (ret != 0) {
        goto cleanup_1;
    }

    /* initialize emac */
    ret = enet_emac_init(&gContext);
    if (ret != 0) {
        goto cleanup_1;
    }

    spin_lock_init(&gContext.lock);

    ret = parse_mac_address_parameter(&gContext);
    if (ret != 0) {
        goto cleanup_1;
    }

    dev = init_etherdev(NULL, 0);
    if (dev == NULL) {
        printk(KERN_ERR "unable to allocate net device structure");
        goto cleanup_1;
    }

    SET_MODULE_OWNER(dev);

    dev->priv = &gContext;
    gContext.net = dev;


    memcpy(dev->dev_addr, gContext.mac_address, 6*sizeof(unsigned char));

    dev->irq                = BCM_IRQ_ENET_CPU_INTR;
    dev->base_addr          = (unsigned long)ENET_BRIDGE;
    dev->open               = enet_open;
    dev->stop               = enet_stop;
    dev->set_mac_address    = enet_set_mac_address;
    dev->tx_timeout         = enet_net_timeout;
    dev->set_multicast_list = enet_set_mcast_list;
    dev->watchdog_timeo     = 2*HZ;
    dev->get_stats          = enet_get_stats;
    dev->hard_start_xmit    = enet_hard_start_xmit;

    write_mac_address(dev);

    PRINTK(KERN_INFO "dev name = %s\n", dev->name);

    ret = request_irq(dev->irq, enet_isr, SA_INTERRUPT,dev->name,dev);
    if (ret != 0) {
        printk(KERN_ERR "request_irq failed (%d)\n", ret);
        goto cleanup_2;
    }


    INIT_LIST_HEAD(&gContext.workers);
    atomic_set(&gContext.reclaimTxDescs, 0);

    register_set_bits(&ENET_BRIDGE_TX_DESC_TABLE[MAX_TX_DESCS-1], DMA_DESC_WRAP);

    for (i = 0; i < gMaxRxDescs; i++) {
        enet_desc_elem_t *elem = NULL;
        elem=(enet_desc_elem_t *)kmalloc(sizeof(enet_desc_elem_t),
                                                GFP_KERNEL);
        memset(elem, 0, sizeof(enet_desc_elem_t));
        elem->descriptor = &ENET_BRIDGE_RX_DESC_TABLE[i];
        gContext.rx_descs[i] = elem;
    }
    register_set_bits(&ENET_BRIDGE_RX_DESC_TABLE[gMaxRxDescs-1], DMA_DESC_WRAP);

    INIT_TQUEUE(&gContext.enet_worker_task, enet_task, NULL);

    INIT_TQUEUE(&gContext.enet_receiver_task, enet_receiver_task_func, NULL);

    gContext.linkState = -1; // force first-time detection
    (void)is_enet_up(ENET_MII_PHYADDR);
    init_timer(&gContext.enet_timer);

    gContext.enet_timer.data        = 0;
    gContext.enet_timer.function    = enet_timer_func;

    enet_rearm_timer_task();


    enet_create_and_assign_rx_skbs();

    gContext.receiver =(enet_worker_t *)kmalloc(sizeof(enet_worker_t),
                                                GFP_KERNEL);
    if (gContext.receiver == NULL) {
        printk(KERN_ERR "enet: unable to allocate receive worker\n");
        ret = -ENOMEM;
        goto cleanup_2;
    }

    memset(gContext.receiver, 0, sizeof(enet_worker_t));
    gContext.receiver->sn = ++gContext.workerSN;

    enet_worker_next_state(gContext.receiver, ENET_WORKER_START_RX_DMA);

    list_add_tail((struct list_head *)gContext.receiver, &gContext.workers);

    enet_make_worker_queue_runnable();
    /* turn on the activity LED */
    *(SUN_TOP_CTRL_PIN_MUX_CTRL_5) |= 0x12000000UL;
    barrier();

    return 0;

cleanup_2:
    enet_cleanup();
cleanup_1:
    return ret;
}

#ifdef COPY_PACKET

#define initialize_rx_buf_for_desc(elem)                                \
    do {                                                                \
        volatile dma_desc_t *desc = (elem)->descriptor;                 \
        (elem)->rxBuffer = (elem)->rxBufferHead;                        \
        unsigned long rsv_bytes = BCM_ENET_ALIGNMENT -                  \
                                  (((unsigned long)((elem)->rxBuffer))% \
                                  BCM_ENET_ALIGNMENT);                  \
        (elem)->rxBuffer+=rsv_bytes;                                    \
        DMA_DESC_FILL_BUFFER_PARAMS((desc),                             \
                                    virt_to_phys((elem)->rxBuffer),     \
                                    (ETH_MAX_MTU_SIZE));                \
        dma_cache_inv((unsigned long)((elem)->rxBuffer),          \
                            (ETH_MAX_MTU_SIZE));                        \
        register_set_bits(&(desc)->length_and_status, DMA_DESC_OWN);    \
    }while (0)


#else

#define associate_rx_skb_with_desc(elem,skb)                            \
    do {                                                                \
        volatile dma_desc_t *desc = (elem)->descriptor;                 \
        unsigned long rsv_bytes = BCM_ENET_ALIGNMENT -                  \
                                  (((unsigned long)(skb)->data)%        \
                                  BCM_ENET_ALIGNMENT);                  \
        /* skb_reserve((skb),2) */;  /* for IP header alignment */      \
        skb_reserve((skb),rsv_bytes) ;  /* for IP header alignment */   \
        DMA_DESC_FILL_BUFFER_PARAMS((desc), virt_to_phys(((skb)->data)),\
                                    (ETH_MAX_MTU_SIZE));                \
        dma_cache_inv((unsigned long)(skb)->data,                 \
                        (ETH_MAX_MTU_SIZE));                            \
        register_set_bits(&(desc)->length_and_status, DMA_DESC_OWN);    \
        (elem)->skb = (skb);                                            \
    }while (0)

#endif /* COPY_PACKET */

static int enet_create_and_assign_rx_skbs()
{
#ifndef COPY_PACKET
    struct sk_buff *skb = NULL;
#endif /* COPY_PACKET */
    int i = 0;

    for (i = 0; i < gMaxRxDescs;i++) {
        enet_desc_elem_t *elem = (enet_desc_elem_t *)gContext.rx_descs[i];

#ifdef COPY_PACKET
        elem->rxBufferHead = (char *)kmalloc(ETH_MAX_MTU_SIZE + BCM_ENET_ALIGNMENT, GFP_KERNEL);
        initialize_rx_buf_for_desc(elem);
#else
        skb = dev_alloc_skb(ETH_MAX_MTU_SIZE + BCM_ENET_ALIGNMENT);
        if (skb == NULL) {
            PRINTK(KERN_ERR "enet: unable to pre-allocate rx skb\n");
            return -ENOMEM;
        }
        associate_rx_skb_with_desc(elem, skb);
#endif /* COPY_PACKET */

    }

    return 0;
}


static int enet_udma_init(enet_device_t *context)
{
    register_set_bits(ENET_DMA_CONTROLLER_CONFIG,IUDMA_CNTRL_CFG_FC_EN | 0x2);

    register_write(ENET_DMA_CONTROLLER_CH1_THRESH_LO, 5);

    register_write(ENET_DMA_CONTROLLER_CH1_THRESH_HI, (gMaxRxDescs & 0xFF));

    register_write(ENET_DMA_CONTROLLER_CH1_FLOW_ALLOC, 
                   (0x1000000UL | (gMaxRxDescs & 0xFF)));

    /* initialize the xmit channel */
    register_set_bits(IUDMA_CHNL_CTL(ENET_DMA_CONTROLLER_TX_CHANNEL),0);

    register_write(IUDMA_CHNL_MAX_BURST(ENET_DMA_CONTROLLER_TX_CHANNEL),
                    DMA_MAX_BURST_LENGTH);

    register_write(IUDMA_CHNL_IRQ_MSK(ENET_DMA_CONTROLLER_TX_CHANNEL),0);

    /* clear any pending interrupts */
    register_set_bits(IUDMA_CHNL_IRQ_STS(ENET_DMA_CONTROLLER_TX_CHANNEL),
                    IUDMA_CHNL_IRQ_STATUS_BDONE | 
                    IUDMA_CHNL_IRQ_STATUS_PDONE | 
                    IUDMA_CHNL_IRQ_STATUS_NOTVID);

    register_write(IUDMA_CHNL_DESC_PTR(ENET_DMA_CONTROLLER_TX_CHANNEL),
                   (unsigned long)ENET_BRIDGE_TX_DESC_TABLE);

    register_set_bits(IUDMA_CHNL_CTL(ENET_DMA_CONTROLLER_RX_CHANNEL),0);
    register_set_bits(IUDMA_CHNL_MAX_BURST(ENET_DMA_CONTROLLER_RX_CHANNEL),
                    DMA_MAX_BURST_LENGTH);

    register_write(IUDMA_CHNL_IRQ_MSK(ENET_DMA_CONTROLLER_RX_CHANNEL),0);

    /* clear any pending interrupts */
    register_set_bits(IUDMA_CHNL_IRQ_STS(ENET_DMA_CONTROLLER_RX_CHANNEL),
                    IUDMA_CHNL_IRQ_STATUS_BDONE | 
                    IUDMA_CHNL_IRQ_STATUS_PDONE | 
                    IUDMA_CHNL_IRQ_STATUS_NOTVID);

    /* interrupt on packet done & notvid */
    register_set_bits(IUDMA_CHNL_IRQ_MSK(ENET_DMA_CONTROLLER_RX_CHANNEL),
                    IUDMA_CHNL_IRQ_MASK_PDONE );

    register_write(IUDMA_CHNL_DESC_PTR(ENET_DMA_CONTROLLER_RX_CHANNEL),
                   (unsigned long)ENET_BRIDGE_RX_DESC_TABLE);

    register_set_bits(ENET_DMA_CONTROLLER_R5K_IRQ_MSK,
                      IUDMA_R5K_IRQ_STATUS_IUDMA);
    return 0;
}

#define MAX_RETRIES         10000

static int disable_emac(void) 
{
    volatile unsigned long val = 0;
    int count = 0;
    do {
        register_read(val, ENET_BRIDGE_CONTROL);
        count++;
        udelay(1000);
    }while ((val & ENET_CONTROL_ED) && (count < MAX_RETRIES));

    if (count >= MAX_RETRIES){
        printk(KERN_ERR "unable to disable emac\n");
        return -1;
    }

    return 0;
}


static int enet_emac_init (enet_device_t *context)
{
    int ret = 0;
    enum mii_config config;

    register_set_bits(ENET_BRIDGE_CONTROL, ENET_CONTROL_ED);
    ret = disable_emac();
    if (ret != 0) {
        return ret;
    }


    register_write(__longreg(&(ENET_BRIDGE->mii_sc)), 
                   (ENET_MII_SC_PRE_EN | 0x01F));
    soft_reset(ENET_MII_PHYADDR);
    config = mii_auto_configure();

    register_set_bits(ENET_BRIDGE_XMIT_CONTROL, ENET_XMIT_CONTROL_FM);

    if (config & MII_FULLDUPLEX) {
        PRINTK(KERN_INFO "enet: discovered a full duplex interface\n");
        register_set_bits(ENET_BRIDGE_XMIT_CONTROL, ENET_XMIT_CONTROL_FD);
    }else {
        PRINTK(KERN_INFO "enet: discovered a half duplex interface\n");
    }
            
    register_write(ENET_BRIDGE_RX_LENGTH, ETH_MAX_MTU_SIZE);
//  register_write(ENET_BRIDGE_TX_LENGTH, ETH_MAX_MTU_SIZE);

    /*
     * watermark = abs(63 - (0.63 * DMA_MAX_BURST_LENGTH))
     * Dont know how the bcm driver code came up with 32, I compute 42,
     * the equation described in the brcm documentation is 
     * watermark <= 511 - (511 % DMA_MAX_BURST_LENGTH)
     */
    register_write(ENET_BRIDGE_XMIT_WATERMARK, 
                   (511 - (63 % ((*IUDMA_CHNL_MAX_BURST(ENET_DMA_CONTROLLER_TX_CHANNEL)) & 0xFF))));

    register_set_bits((__longreg(&((ENET_BRIDGE)->mib_ctrl))), 1);

    /* mask all EMAC interrupts */
    register_write(ENET_BRIDGE_INT_MASK, 0);

    return 0;
}

static int parse_mac_address_parameter(enet_device_t *context)
{
    unsigned char *ptr = mac;
    int i = 0;

    if (ptr == NULL) {
        printk(KERN_ERR "MAC address not passed in as module parameter\n");
        return -1;
    }

    if ( (mac[0] == 0) && (mac[1] == 0) && (mac[2] == 0) &&
         (mac[3] == 0) && (mac[4] == 0) && (mac[5] == 0)) {
        printk(KERN_ERR "ENET: Mac address not configured\n");
        return -1;
    }

    for (i = 0;  (ptr != NULL) && (*ptr != '\0') && (i < 6);) {
        char *endptr;
        if ('0' == *ptr) {
            ptr++;
            if (ptr != NULL  &&  (*ptr == 'x' || *ptr == 'X')) {
                    ptr++;
                    if (ptr == NULL || *ptr == '\0') {
                        break;
                    } else {
                        continue;
                    }
            }
        }
            
        int mac_byte = (int)simple_strtoul((char *)ptr, &endptr, 16);
        context->mac_address[i++] = (unsigned char)mac_byte;
        ptr = endptr;
        if ((ptr != NULL) && (*ptr != '\0')) {
            ptr++;
        }
    }

    printk("MAC initialized OK %2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X\n", 
           context->mac_address[0], context->mac_address[1], 
           context->mac_address[2], context->mac_address[3], 
           context->mac_address[4], context->mac_address[5]);

    return 0;
}
#define MAKE_SHORT_WORD(x)                                               \
    ((x)[1] & 0xFF) | (((x)[0] & 0xFF) << 8)

#define MAKE_LONG_WORD(x)                                                \
    (( ((x)[3] & 0xFF))     | (((x)[2] & 0xFF) << 8) |                   \
    ( ((x)[1] & 0xFF) << 16)| (((x)[0] & 0xFF) << 24))

static int pm_reg_write(int index, unsigned char *addr, int valid)
{
    volatile pm_data_t *pmdata = ENET_BRIDGE_PM_DATA(index);

    unsigned short high = MAKE_SHORT_WORD(addr);
    unsigned long  low  = MAKE_LONG_WORD(addr+2);

    register_write(&(pmdata->low),low);    

    register_set(&(pmdata->high),                           
                 PM_HIGH_DATA_H_MASK,                       
                 PM_HIGH_DATA_H_SHIFT,                      
                 high);                                     
    register_set(&(pmdata->high),                           
                 PM_HIGH_VALID_MASK,                        
                 PM_HIGH_VALID_SHIFT,                       
                 valid ? 1 : 0);                            

    return 0;
}


static int is_same_address(volatile pm_data_t *pmreg, unsigned char *addr)
{
    unsigned short high;
    unsigned long  low;

    register_read(high, &(pmreg->high));
    register_read(low, &(pmreg->low));

    if ((memcmp(&high, addr, sizeof(high)) == 0) &&
        (memcmp(&low, addr+2, sizeof(low)) == 0)) {
        return 1;
    }

    return 0;
}


static int perfectmatch_update(enet_device_t *context, unsigned char*addr,
                               int fValid)
{
    int i = 0;
    unsigned int aged_ref;
    int aged_entry;

    /* check if this address is in used */
    for (i = 0; i < MAX_PM_REGS; i++) {

        volatile pm_data_t *pmreg = ENET_BRIDGE_PM_DATA(i);

        if (IS_PM_VALID(pmreg)) {

            if (is_same_address(pmreg, addr)) {
                if (fValid == 0) {
                    /* clear the valid bit in PM register */
                    pm_reg_write(i, addr, fValid);
                } else {
                    context->pm_reg_entry[i]++;
                }
                return 0;
            }
        }
    }
    if (fValid == 1) {
        for (i = 0; i < MAX_PM_REGS; i++) {
            /* find an available entry for write pm address */
            if (IS_PM_VALID(ENET_BRIDGE_PM_DATA(i)) == 0) {
                break;
            }
        }

        if (i == MAX_PM_REGS) {
            aged_ref = 0xFFFFFFFF;
            aged_entry = 0;
            /* aged out an entry */
            for (i = 0; i < MAX_PM_REGS; i++) {
                if (context->pm_reg_entry[i] < aged_ref) {
                    aged_ref = context->pm_reg_entry[i];
                    aged_entry = i;
                }
            }
            i = aged_entry;
        }
        /* find a empty entry and add the address */
        pm_reg_write(i, addr, fValid);

        /* update the pDevCtrl->pmAddr */
        context->pm_reg_entry[i] = 1;
    }

    return 0;
}

static void perfectmatch_clean(enet_device_t *context, unsigned char *addr)
{
    int i;

    for (i = 0; i < MAX_PM_REGS; i++) {
        volatile pm_data_t *pmreg = ENET_BRIDGE_PM_DATA(i);
        if (IS_PM_VALID(pmreg)) {
            if (!is_same_address(pmreg, addr)) {
                pm_reg_write(i, addr, 0);
            }
        }
    }
}

static int write_mac_address(struct net_device* net)
{
    enet_device_t *context = (enet_device_t *)net->priv;
    int ret = 0;
    volatile unsigned long val = 0;
    int i;

    if (context == NULL) {
        return -1;
    }

    register_read(val, ENET_BRIDGE_CONTROL);

    ret = disable_emac();
    if (ret != 0) {
        return ret;
    }

    perfectmatch_update(context, net->dev_addr, 1);

    if (val & ENET_CONTROL_EE) {
        register_set_bits(ENET_BRIDGE_CONTROL, ENET_CONTROL_EE);
    }
    for (i = 0; i< MAX_PM_REGS; i++) {
        PRINTK("PM%d_DATA_LO = 0x%08X PM%d_DATA_HI = 0x%08X\n", i, (unsigned int)ENET_BRIDGE_PM_DATA(i)->low,i, (unsigned int)ENET_BRIDGE_PM_DATA(i)->high);
    }

    return 0;
}

static int enet_set_mac_address (struct net_device *net, void *p)
{
    struct sockaddr *addr = p;

    if (netif_running(net)) {
        return -EBUSY;
    }

    memcpy(net->dev_addr, addr->sa_data, net->addr_len);

    write_mac_address(net);

    return 0;
}

static int enet_open (struct net_device *net)
{
    enet_device_t *context = (enet_device_t *)net->priv;
    volatile unsigned long flags;
    int ret = 0;

    if (context == NULL) {
        return -ENODEV;
    }


    MOD_INC_USE_COUNT;

    spin_lock_irqsave(&context->lock, flags);

    ret = disable_emac();
    if (ret != 0) {
        MOD_DEC_USE_COUNT;
        spin_unlock_irqrestore(&context->lock, flags);
        return -EIO;
    }

    register_set_bits(ENET_BRIDGE_CONTROL, ENET_CONTROL_EE);

    /* enable flow control */
    register_set_bits(ENET_BRIDGE_RX_CONFIG,RX_CONFIG_ENFLOW);

    register_set_bits(IUDMA_CHNL_CTL(ENET_DMA_CONTROLLER_RX_CHANNEL), 
                      IUDMA_CHNL_CTL_KICKED);

    register_set_bits(IUDMA_CHNL_IRQ_MSK(ENET_DMA_CONTROLLER_RX_CHANNEL), 
                      IUDMA_CHNL_IRQ_MASK_PDONE);

    /* TODO timer stuff */
    netif_start_queue(net);

    spin_unlock_irqrestore(&context->lock, flags);

    return 0;
}

static int enet_stop (struct net_device *net)
{
    enet_device_t *context = (enet_device_t *)net->priv;
    volatile unsigned long flags;
    int ret = 0;

    if (context == NULL){
        return -ENODEV;
    }

    netif_stop_queue(net);

    spin_lock_irqsave(&context->lock, flags);

    /* disable emac */
    register_clear(ENET_BRIDGE_CONTROL, ENET_CONTROL_EE_MASK);

    ret = disable_emac();
    if (ret != 0) {
        spin_unlock_irqrestore(&context->lock, flags);
        return -EIO;
    }


    /* stop RX channel DMA */
    register_clear(IUDMA_CHNL_CTL(ENET_DMA_CONTROLLER_RX_CHANNEL), 
                   IUDMA_CHNL_CTL_KICKED_MASK);

    register_clear(IUDMA_CHNL_IRQ_MSK(ENET_DMA_CONTROLLER_RX_CHANNEL), 
                   IUDMA_CHNL_IRQ_MASK_PDONE_MASK);

    register_clear(IUDMA_CHNL_IRQ_MSK(ENET_DMA_CONTROLLER_RX_CHANNEL), 
                   IUDMA_CHNL_IRQ_MASK_NOTVID_MASK);

    spin_unlock_irqrestore(&context->lock, flags);
    MOD_DEC_USE_COUNT;

    return 0;
}

static void enet_net_timeout (struct net_device *net)
{
    if (net == NULL) {
        printk(KERN_ERR "enet: NULL net_device pointer in %s\n", 
               __FUNCTION__);
        return;
    }

    net->trans_start = jiffies;

    netif_wake_queue(net);
}

static void enet_set_mcast_list (struct net_device *net)
{
    unsigned int prom = 0;
    unsigned int multi = 0;
    if (net == NULL) {
        printk(KERN_ERR "enet: NULL net_device pointer in %s\n", __FUNCTION__);
        return;
    }

    /* clear filter list (except MAC address) */
    perfectmatch_clean((enet_device_t *)net->priv, net->dev_addr);

    /* promiscuous mode */
    if (net->flags & IFF_PROMISC) {
        prom = 1;
    /* all multicast packets, or too many addresses */
    } else if ((net->flags & IFF_ALLMULTI) ||
               (net->mc_count > (MAX_PM_REGS - 1))) {
        multi = 1;
    /* no filter list */
    } else if (net->mc_count == 0) {
    /* add to filter list */
    } else {
        struct dev_mc_list *mc_list = net->mc_list;
        int i;

        for (i = 0; i < net->mc_count; i++) {
            perfectmatch_update((enet_device_t *)net->priv,
                                mc_list->dmi_addr, 1);
            mc_list = mc_list->next;
        }
    }

    register_set(ENET_BRIDGE_RX_CONFIG,RX_CONFIG_PROM_MASK, 
                 RX_CONFIG_PROM_SHIFT, prom);

    register_set(ENET_BRIDGE_RX_CONFIG,RX_CONFIG_ACCMULT_MASK,
                 RX_CONFIG_ACCMULT_SHIFT, multi);

}

static struct net_device_stats * enet_get_stats(struct net_device *net) 
{
    enet_device_t *context = NULL;

    if (net == NULL) {
        printk(KERN_ERR "enet: NULL net_device pointer in %s\n", 
               __FUNCTION__);
        return NULL;
    }

    context = (enet_device_t *)net->priv;
    if (net == NULL) {
        printk(KERN_ERR "enet: NULL context pointer in %s\n", 
               __FUNCTION__);
        return NULL;
    }

    return &(context->stats);
}


#define next_tx_desc(i) (((i) + 1)%MAX_TX_DESCS)

static int enet_get_next_tx_desc(int *idx)
{
    int iter = 0;
    int pending = 0;
    *idx = -1;

    for (iter = gContext.curTxIdx; ; iter = next_tx_desc(iter)) {
        volatile dma_desc_t *desc = &ENET_BRIDGE_TX_DESC_TABLE[iter];
        if (is_desc_free(desc))  {
            if (iter == *idx) break;
            if (*idx == -1) {
                DMA_DESC_FILL_BUFFER_PARAMS(desc, 0, 0);
                *idx = iter;
                gContext.curTxIdx = next_tx_desc(iter);
            }
            if (NULL != gContext.tx_descs[iter]) {
                dev_kfree_skb_any(gContext.tx_descs[iter]);
                gContext.tx_descs[iter] = NULL;
            }
        } else {
            ++pending;
            break;
        }
    }


    if (pending > (MAX_TX_DESCS/2)) {
        return 1;
    }

    return 0;

}

static int enet_hard_start_xmit(struct sk_buff *skb, struct net_device *net)
{
    int pending = 0;
    int idx = 0;
    volatile dma_desc_t *desc = NULL;

    netif_stop_queue(net);

    net->trans_start = jiffies;

    pending =  enet_get_next_tx_desc(&idx);
    if (idx < 0 ) {
        dev_kfree_skb_any(skb);
        return NET_XMIT_SUCCESS;
    }

    desc = &ENET_BRIDGE_TX_DESC_TABLE[idx];
    gContext.tx_descs[idx] = skb;

    DMA_DESC_FILL_BUFFER_PARAMS(desc,virt_to_phys(skb->data), 
                                (skb->len > ETH_ZLEN) ? 
                                skb->len : ETH_ZLEN);
    register_set_bits(&(desc->length_and_status), DMA_DESC_PS_APPEND_CRC);

    dma_cache_wback_inv((unsigned long)skb->data, skb->len);

    register_write(IUDMA_CHNL_IRQ_MSK(ENET_DMA_CONTROLLER_TX_CHANNEL),
                      0);
    register_set_bits(desc, DMA_DESC_OWN | DMA_DESC_SOP | DMA_DESC_EOP);

    register_set_bits(IUDMA_CHNL_CTL(ENET_DMA_CONTROLLER_TX_CHANNEL), 
                      IUDMA_CHNL_CTL_KICKED);
    if (!pending) {
        netif_start_queue(net);
    } else {
        atomic_set(&gContext.reclaimTxDescs, 1);
    }

    gContext.stats.tx_bytes += ((skb->len < ETH_ZLEN) ? ETH_ZLEN : skb->len);
    gContext.stats.tx_bytes += 4;
    gContext.stats.tx_packets++;

    return NET_XMIT_SUCCESS;
}

volatile unsigned long gRxInProgress = 0;

static void enet_isr(int irq, void *dev_id, struct pt_regs *regs)
{
    volatile unsigned short txStatus;
    volatile unsigned short rxStatus;

    register_read(txStatus,IUDMA_CHNL_IRQ_STS(ENET_DMA_CONTROLLER_TX_CHANNEL));
    register_read(rxStatus,IUDMA_CHNL_IRQ_STS(ENET_DMA_CONTROLLER_RX_CHANNEL));

    if (txStatus) {
        register_set_bits(IUDMA_CHNL_IRQ_STS(ENET_DMA_CONTROLLER_TX_CHANNEL),
                          txStatus);
    }

    if (rxStatus) {
        register_set_bits(IUDMA_CHNL_IRQ_STS(ENET_DMA_CONTROLLER_RX_CHANNEL),
                          rxStatus);
      if (!(test_bit(0, &gRxInProgress))) {
          gContext.receiver->go = 1;
          register_set_bits(IUDMA_CHNL_CTL(ENET_DMA_CONTROLLER_RX_CHANNEL), 
                            IUDMA_CHNL_CTL_KICKED);
      }
    }

    enet_make_worker_queue_runnable();
}


/**************************************************************************
 *                  enet task and worker state machine                    *
 *                                                                        *
 *************************************************************************/

static void enet_timer_func(unsigned long data)
{
    (void)data;

    (void)is_enet_up(ENET_MII_PHYADDR);
    if (atomic_read(&gContext.reclaimTxDescs)) {
        int i = 0;
        int count = 0;
        for (i = 0 ; i < MAX_TX_DESCS; ++i) {
            volatile dma_desc_t *desc = &ENET_BRIDGE_TX_DESC_TABLE[i];
            if (is_desc_free(desc)) {
                ++count;
            }
        }

        if (count > 8) {
            atomic_set(&gContext.reclaimTxDescs, 0);
            netif_start_queue(gContext.net);
        }
    }
    enet_rearm_timer_task();

}

static unsigned long main_running = 0;

static void enet_main(void)
{
    int done;

    do {
        done = 1;
        done &= enet_run_workers();
    }while (!done);
    clear_bit(0, &main_running);
}
    

static __inline__ void run_main(void)
{
    if (!test_and_set_bit(0, &main_running)) {
        enet_main();
    }
}

static void enet_task(void *data)
{
    (void)data;
    run_main();
}


static void enet_receiver_task_func (void *data)
{
    unsigned long curIdx = (unsigned long)data;
    enet_desc_elem_t *elem = NULL;
    volatile dma_desc_t *desc = NULL;

    register_clear(ENET_DMA_CONTROLLER_R5K_IRQ_MSK, 
                   IUDMA_R5K_IRQ_STATUS_IUDMA_MASK);
    curIdx = enet_do_packet_receive(curIdx);
    gContext.receiver->priv = curIdx;
    elem = (enet_desc_elem_t *)gContext.rx_descs[curIdx];
    desc = elem->descriptor;
    
    //if (desc->length_and_status
    /* go back and wait for next rx interrupt */
    clear_bit(0, &gRxInProgress);
    enet_worker_next_state(gContext.receiver, ENET_WORKER_START_RX_DMA);
    register_set_bits(ENET_DMA_CONTROLLER_R5K_IRQ_MSK,
                      IUDMA_R5K_IRQ_STATUS_IUDMA);
    register_set_bits(IUDMA_CHNL_CTL(ENET_DMA_CONTROLLER_RX_CHANNEL), 
                      IUDMA_CHNL_CTL_KICKED);

    enet_make_worker_queue_runnable();
}

static int enet_run_workers()
{
    int done = 1;
    struct list_head *this, *next;
    enet_worker_t *worker;

    PRINTK(KERN_INFO "enet: running worker queue\n");
    list_for_each_safe(this, next, &gContext.workers) {
        worker = (enet_worker_t *)this;
        if (worker->go) {
            PRINTK(KERN_INFO "enet: worker %d go!\n", worker->sn);
            worker->go = 0;

            switch (worker->state) {
              case ENET_WORKER_START_RX_DMA:
                enet_worker_start_rx_dma(worker);
                break;

              case ENET_WORKER_FINISH_RX_DMA:
                enet_worker_finish_rx_dma(worker);
                break;

              default:
                break;
            }
        }else {
            PRINTK(KERN_INFO "enet: worker %d nogo\n", worker->sn);
        }
    }

    list_for_each (this, &gContext.workers) {
        worker = (enet_worker_t *)this;
        done &= ! worker->go;
    }

    PRINTK(KERN_INFO "enet: queue runner says %sdone\n",done ? "" : "not ");
                    
    return done;
}

#define reset_emac_on_errors()                                          \
    do {                                                                \
        volatile unsigned long flags;                                   \
        spin_lock_irqsave(&gContext.lock, flags);                       \
        gNumberOfOverflows = 0;                                         \
        spin_unlock_irqrestore(&gContext.lock, flags);                  \
        register_set_bits(ENET_BRIDGE_CONTROL, ENET_CONTROL_ED);        \
        (void)disable_emac();                                           \
        register_set_bits(ENET_BRIDGE_CONTROL, ENET_CONTROL_EE);        \
    }while(0)


static void enet_worker_start_rx_dma(enet_worker_t *worker) 
{
    PRINTK(KERN_INFO "enet: worker %d starting rx dma\n", worker->sn);

    register_set_bits(IUDMA_CHNL_IRQ_MSK(ENET_DMA_CONTROLLER_RX_CHANNEL),
                      IUDMA_CHNL_IRQ_MASK_PDONE );

    worker->state = ENET_WORKER_FINISH_RX_DMA;
}


#define OVERFLOWTHRESHOLD       8

#define is_desc_error(desc)                                             \
    (((desc)->length_and_status & (DMA_DESC_EOP)) &&                    \
     ((desc)->length_and_status & (DMA_DESC_PS_OV | DMA_DESC_PS_CRC |   \
                                  DMA_DESC_PS_NO | DMA_DESC_PS_LG  |    \
                                  DMA_DESC_PS_RXERR)))

static void enet_worker_finish_rx_dma(enet_worker_t *worker)
{
    PRINTK(KERN_INFO "enet: worker %d finish rx dma\n", worker->sn);

    if (!(test_and_set_bit(0, &gRxInProgress))) {
#ifdef INCONVENIENTLY_FRIENDLY_RX
        unsigned long curIdx = worker->priv;
        gContext.enet_receiver_task.data = (void *)curIdx;
        queue_task(&gContext.enet_receiver_task, &tq_immediate);
        mark_bh(IMMEDIATE_BH);
#endif /* INCONVENIENTLY_FRIENDLY_RX */
        enet_receiver_task_func ((void *)(worker->priv));
    }
}

static unsigned long enet_do_packet_receive(unsigned long curIdx)
{
#ifndef COPY_PACKET
    struct sk_buff *skb = NULL;
#endif /* COPY_PACKET */
    unsigned long plen = 0;
    unsigned long blen = 0;
    unsigned int perr = 0;
    unsigned int continueReadingUntilEOP = 0;
    int i = 0;
    int fIncIndex = 1;
    unsigned long nPackets = 0;

    for (i = curIdx;; i= ((i+1)%gMaxRxDescs)) {
        enet_desc_elem_t *elem = (enet_desc_elem_t *)gContext.rx_descs[i];
        volatile dma_desc_t *desc = elem->descriptor;
        struct sk_buff *received_skb = NULL;
        unsigned int eop = 0;

        if ((desc->length_and_status & DMA_DESC_OWN) || (desc->buffer == 0)) {
            if (gNumberOfOverflows > OVERFLOWTHRESHOLD) {
                reset_emac_on_errors();
            }
            /* we would have caught up with the DMA engine */
            /* the worker should now point to this descriptor element */
            fIncIndex = 0;
            break;
        }

        if (continueReadingUntilEOP != 0) {
            plen += (desc->length_and_status >> 16) & 0xFFFUL;;

            gContext.stats.rx_dropped++;
            gContext.stats.rx_errors++;
            if (desc->length_and_status & DMA_DESC_PS_OV) {
                gNumberOfOverflows++;
            }

            if ((desc->length_and_status & DMA_DESC_EOP) == 0) {
#ifdef COPY_PACKET
                initialize_rx_buf_for_desc(elem);
#else
                DMA_DESC_FILL_BUFFER_PARAMS(desc, desc->buffer, ETH_MAX_MTU_SIZE);
                register_set_bits(&(desc)->length_and_status, DMA_DESC_OWN);
#endif /* COPY_PACKET */
                perr++;
                continue;
            }else {
                continueReadingUntilEOP = 0;
                perr = 0;
            }
        }

        if (is_desc_error(desc)) {
            /* discard packets with errors */
            gContext.stats.rx_dropped++;
            gContext.stats.rx_errors++;
            plen += (desc->length_and_status >> 16) & 0xFFFUL;;

            continueReadingUntilEOP = 1;

            if (desc->length_and_status & DMA_DESC_PS_OV) {
                gNumberOfOverflows++;
            }

            if ( (desc->length_and_status & DMA_DESC_PS_LG)){
                reset_emac_on_errors();
            }

            if ( plen > 2048) {
                /* for oversized packets, the DMA_DESC_PS_LG bit may not be set
                 * by the hardware. We handle this case separately and reset
                 * the controller.
                 */
                if (gNumberOfOverflows > OVERFLOWTHRESHOLD) {
                    reset_emac_on_errors();
                }
            }

#ifdef COPY_PACKET
            initialize_rx_buf_for_desc(elem);
#else
            DMA_DESC_FILL_BUFFER_PARAMS(desc, desc->buffer, ETH_MAX_MTU_SIZE);
            register_set_bits(&(desc)->length_and_status, DMA_DESC_OWN);
#endif /* COPY_PACKET */

            perr++;
            continue;
        }


        if ((desc->length_and_status & DMA_DESC_EOP) != 0) {
            eop = 1;
            ++nPackets;
        } 

        perr = 0;
        gNumberOfOverflows = 0;
        blen = (desc->length_and_status>>16) & 0xFFFUL;
#ifdef COPY_PACKET
        blen-=4;
        elem->rxBuffer = phys_to_virt((unsigned long)elem->rxBuffer);
        received_skb = dev_alloc_skb(blen + 2);
        if (received_skb == NULL) {
            fIncIndex = 0;
            break;
        }
        skb_reserve(received_skb, 2);
        received_skb->dev = gContext.net;
        eth_copy_and_sum(received_skb, elem->rxBuffer, (int)blen, 0);
        skb_put(received_skb, blen);

        initialize_rx_buf_for_desc(elem);
#else
        elem->skb->data = phys_to_virt((unsigned long)(elem->skb->data));
        skb_put(elem->skb, (blen -4) ); /* trim the ethernet CRC */
        elem->skb->dev = gContext.net;

        received_skb = elem->skb;
        elem->skb = NULL;

        skb = dev_alloc_skb(ETH_MAX_MTU_SIZE + BCM_ENET_ALIGNMENT);
        if (skb == NULL) {
            PRINTK(KERN_ERR "enet: unable to pre-allocate rx skb\n");
            panic("enet: unable to allocate rx skbuff\n");
        }
        associate_rx_skb_with_desc(elem, skb);
#endif /* COPY_PACKET */

        received_skb->protocol = eth_type_trans(received_skb, gContext.net);

//#define DUMP_DATA
#ifdef DUMP_DATA
        int j = 0;
        unsigned char *pBuf = (unsigned char *)received_skb->head;
        printk("data (%d bytes)" , blen);
        for (j = 0; j< blen; j++) {
            printk("0x%02X ", pBuf[j]);
        }
        printk("\n");
#endif /* DUMP_DATA */


        /* notify the kernel */
        netif_rx(received_skb);
        gContext.net->last_rx = jiffies;
        gContext.stats.rx_bytes+=blen;
        if (eop) {
            gContext.stats.rx_packets++;
        }
    }


    if (fIncIndex) {
        curIdx = (i+1)% gMaxRxDescs;
    }else {
        curIdx = i;
    }

    return curIdx;
}

static void __exit enet_exit( void )
{
    enet_cleanup();
}

module_init( enet_init );
module_exit (enet_exit);
EXPORT_NO_SYMBOLS;
