/*
 * net/ipv4/fasttcp.c
 *
 * Copyright (C) 2005 TiVo Inc. All Rights Reserved.
 *
 * Written by Shane Nay (snay@tivo.com)
 *
 * As of this writing (8/22/2005) we're currently only modifying the
 * entry point into the TCP stack.  Eventually, we'll subsume much
 * of the stack directly for a fast path implementation.
 *
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>

#include <net/icmp.h>
#include <net/tcp.h>

#include <linux/fasttcp.h>
#include <asm/io.h>

#include <asm/uaccess.h>
#include <asm/ioctls.h>

void tcp_fast_skb_destructor(struct sk_buff* skb) {
	struct fast_tcp_pkt* ftcp= (struct fast_tcp_pkt*)
		(((unsigned int)skb) & (~0x7ff));
	/*
	 * All we really do is flush out the packet.
	 * Well, that and we mark it as being transmitted.
	 * Err, and .., okay, if its the last one in a
	 * "sequence" we wake up the user task that's
	 * waiting on completion of the sequence.
	 * Hmm, and well..., psych!, that's it.
	 */
	ftcp->flags |= FAST_TCP_FLAG_TRANSMIT_FINISHED;
	if((ftcp->flags & FAST_TCP_FLAG_MORE_PACKETS) == 0) {
//		printk("sent last %lu\n", jiffies);
		wake_up(&ftcp->txfin);
	}
	dma_cache_wback_inv((unsigned int)ftcp, 0x800);
}

static inline struct sk_buff *tcp_mk_fast_skb(struct sock *sk,
				      struct fast_tcp_pkt* ftcp) {
	struct sk_buff* skb=&ftcp->skb;
	u8* data;
	unsigned int size = FAST_TCP_MAX_MSS + FAST_TCP_MAX_HEADER - sizeof(struct skb_shared_info);
	data = ftcp->pktdata;
	size = SKB_DATA_ALIGN(size);
	skb->truesize=size + sizeof(struct sk_buff);
	
	skb->head = data;
	skb->data = data;
	skb->tail = data;
	skb->end = data + size;
	skb->len=0;
	skb->cloned = 0;
	skb->data_len = 0;
	skb->nocache = 1;

	atomic_set(&skb->users, 1);
	atomic_set(&(skb_shinfo(skb)->dataref), 1);
	skb_shinfo(skb)->nr_frags = 0;
	skb_shinfo(skb)->frag_list = NULL;

	skb->destructor=tcp_fast_skb_destructor;
	skb_reserve(skb, FAST_TCP_MAX_HEADER);
	return skb;
}

inline int find_user_page(struct page **page, unsigned int uaddr) {
	pgd_t *pgd;
	pmd_t *pmd;
	int ret = -EINVAL;
	struct mm_struct *mm = current->mm;
	spin_lock(&mm->page_table_lock);
	pgd = pgd_offset(current->mm, uaddr);
	pmd = pmd_offset(pgd, uaddr);
	if (pmd) {
		pte_t *pte = pte_offset(pmd, uaddr);
		if( pte && pte_present(*pte)) {
			*page = pte_page(*pte);
			ret = 0;
		}
	}
	spin_unlock(&mm->page_table_lock);
	return ret;
}

int do_tcp_wait_fast_skb(struct sock* sk, unsigned int uaddr) {
	struct page *page;
	struct fast_tcp_pkt* ftcp = NULL;
	int ret = find_user_page(&page, uaddr);
	ftcp=(struct fast_tcp_pkt*)(page_address(page) + (uaddr & 0xfff));
	if(ftcp->flags & FAST_TCP_FLAG_MORE_PACKETS) {
		printk("FAST TCP: Waiting on other than last packet"
		       "not allowed!\n");
		return -EINVAL;
	}
//	printk("Wait on 0x%x\n", (unsigned int)ftcp);
	ret= wait_event_interruptible(ftcp->txfin,
		  (ftcp->flags & FAST_TCP_FLAG_TRANSMIT_FINISHED));

	/* Invalidate the flags field...since we touched it */
	dma_cache_wback_inv((unsigned int)ftcp, 4);
	return ret;
}

ssize_t do_tcp_send_fast_skbs(struct sock *sk, unsigned int uaddr) {
	struct page *page = NULL;
	struct fast_tcp_pkt* ftcp = NULL;
	struct tcp_opt *tp = &(sk->tp_pinfo.af_tcp);
	ssize_t copied=0;
	int ret;

	/* Make sure uaddr is on 0x1000 boundary */
	if(uaddr & 0xfff) {
		printk("FAST TCP: SKB's must start on a 0x1000 boundary!\n");
		return -EINVAL;
	}
//	printk("dtsfb start %lu\n", jiffies);

	lock_sock(sk);

	do {
		struct sk_buff* skb;
		if((uaddr & 0x800) == 0) {
			/* We're on a page boundary, need to pin down */
			__u32 UVflags;
			__get_user(UVflags, (__u32*)uaddr);

			dma_cache_wback_inv(uaddr, 0x1000);
			
			if(!(UVflags & FAST_TCP_FLAG_PAGE_LOCKED)) {
				down_read(&current->mm->mmap_sem);
				ret = get_user_pages(current, current->mm, uaddr,
						     1, 1, 1, &page, NULL);
				up_read(&current->mm->mmap_sem);
				if(ret <= 0) {
					printk("FAST TCP: get_user_pages failed for "
				       "address %08x\n", uaddr);
					goto out;
				}
			} else {
				ret = find_user_page(&page, uaddr);
				if(ret < 0) {
					printk("FAST TCP: find_user_pages failed for "
					       "address %08x\n", uaddr);
					goto out;
				}
			}

			/* Convert to KSEG0 addr */
			
			ftcp=(struct fast_tcp_pkt*)page_address(page);
			
		} else {
			/* We're not on a page boundary, so we need to
			   advance to the next structure */

			ftcp++;
		}
	
		/* Turn it into a fast_skb */
		ftcp->flags &= ~(FAST_TCP_FLAG_TRANSMIT_FINISHED);
		ftcp->flags |= FAST_TCP_FLAG_PAGE_LOCKED;
		skb=tcp_mk_fast_skb(sk, ftcp);

		if(skb == NULL) {
			printk("FAST TCP: Unable to make fast SKB\n");
			ret= -EINVAL;
			goto out;
		}

		skb_entail(sk, tp, skb);
		skb_put(skb, ftcp->size);

		if(!(ftcp->flags & FAST_TCP_FLAG_USERCALC_CRC_PARTIAL)) {
			/* Kernel has to calculate the partial CRC */
			skb->csum=csum_partial(skb->data, ftcp->size, 0);
		} else {
			skb->csum=ftcp->crc_partial;
		}
		/* Put in the transmit pipeline */

		tp->write_seq += ftcp->size;
		TCP_SKB_CB(skb)->end_seq += ftcp->size;
#if 0
		printk("Frame size %d put in TX pathway, uadder 0x%08x\n", ftcp->size, uaddr);
#endif


		copied += ftcp->size;
		uaddr+=0x800;
	} while (ftcp->flags & FAST_TCP_FLAG_MORE_PACKETS);
	init_waitqueue_head(&ftcp->txfin);
     
	if (copied)
		tcp_push( sk, tp, 0, tcp_current_mss(sk), tp->nonagle);
//	printk("dtsfb exit %lu\n", jiffies);
	release_sock(sk);
	return copied;
out:
	release_sock(sk);
	return ret;
}

int do_tcp_release_fast_skb(struct sock *sk, unsigned int uaddr) {
	struct page *page = NULL;
	struct fast_tcp_pkt* ftcp = NULL;

	 /* Make sure uaddr is on 0x1000 boundary */
	if(uaddr & 0xfff) {
		printk("FAST TCP: SKB's must start on a 0x1000 boundary!\n");
		return -EINVAL;
	}

	do {
		if((uaddr & 0x800) == 0) {
			/* We're on a page boundary, need to unpin */
			__u32 UVflags;
			__get_user(UVflags, (__u32*)uaddr);

			dma_cache_wback_inv(uaddr, 0x1000);
			
			if(UVflags & FAST_TCP_FLAG_PAGE_LOCKED) {
				if(find_user_page(&page, uaddr) < 0) {
					printk("FAST TCP: find_user_page failed for "
					       "address %08x\n", uaddr);
					return -EINVAL;
				}
				
				page_cache_release(page);
			} else {
				printk("FAST TCP: Passed unlocked page to"
				       "ioctl\n");
				return -EINVAL;
			}
			
		} else {
			/* We're not on a page boundary, so we need to
			   advance to the next structure */

			ftcp++;
		}

		uaddr+=0x800;
	} while (ftcp->flags & FAST_TCP_FLAG_MORE_PACKETS);
	return 0;
}
